"""
The module :mod:`unpoly.up` exposes the main :class:`.Unpoly` class which is used to communicate
with Unpoly via HTTP headers. All the information of the
`Unpoly server protocol <https://unpoly.com/up.protocol>`_ is exposed and easily accessable.

An example usage could look like this (`request.up` is enabled via the :ref:`Django Integration <django>`):

.. code-block:: python

    def create_user(request):
        # Set a nice <title> for unpoly
        request.up.set_title("Create user | MyDomain.com")
        if request.method == "POST":
            form = UserForm(request.POST)
        else:
            form = UserForm()
        # Do not save the form if unpoly is just trying
        # to validate a form field
        if form.is_valid() and not request.up.validate:
            instance = form.save()
            # Tell unpoly that the user was created successfully
            request.up.emit("user:created", {"id": instance.pk})
            # and also close the layer
            request.up.layer.accept()
        return render(request, "template.html", {"form": form})
"""
from datetime import datetime
from typing import Any, Mapping, Optional
from urllib.parse import parse_qs, urlencode

from .options import Options

try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property


def header_to_opt(header: str) -> str:
    return header[5:].lower().replace("-", "_")


def opt_to_header(opt: str) -> str:
    parts = map(lambda x: x.capitalize(), opt.split("_"))
    return f"X-Up-{'-'.join(parts)}"


def param_to_opt(param: str) -> str:
    return param[4:]


def opt_to_param(opt: str) -> str:
    return f"_up_{opt}"


class Layer:
    def __init__(self, unpoly: "Unpoly", mode: str, context: str):
        self.unpoly = unpoly
        self.mode = mode or "root"  #: Current layer mode (`X-Up[-Fail]-Mode`).
        self.context = context  #: Current layer context (`X-Up[-Fail]-Context`).

    @property
    def is_root(self) -> bool:
        """Returns whether this is the root layer."""
        return self.mode == "root"

    @property
    def is_overlay(self) -> bool:
        """Returns whether this is an overlay layer."""
        return not self.is_root

    def emit(self, type: str, options: Mapping[str, Any]):
        """Emit events to the frontend (`X-Up-Events`).

        Similar to :meth:`.Unpoly.emit` but emits on the current layer.
        """
        options = options or {}
        self.unpoly.emit(type, dict(layer="current", **options))

    def accept(self, value: Optional[Mapping[str, Any]] = None):
        """Accept the current layer (`X-Up-Accept-Layer`).

        An optional value can be provided to be passed back to the client."""
        # assert self.is_overlay -- the client simply ignores this on the root layer
        self.unpoly.options.accept_layer = value or {}

    def dismiss(self, value: Optional[Mapping[str, Any]] = None):
        """Accept the current layer (`X-Up-Dismiss-Layer`).

        An optional value can be provided to be passed back to the client."""
        # assert self.is_overlay -- the client simply ignores this on the root layer
        self.unpoly.options.dismiss_layer = value or {}


class Cache:
    def __init__(self, unpoly):
        self.unpoly = unpoly

    def clear(self, pattern: str = "*"):
        """Tell the client to remove all caches matching the pattern (`X-Up-Clear-Cache`)."""
        self.unpoly.options.clear_cache = pattern

    def keep(self):
        """Tell the client to keep the cache."""
        self.clear("false")  # this is intentional


class Unpoly:
    """The main entrypoint for communication with Unpoly"""

    def __init__(self, adapter):
        self.adapter = adapter

    @cached_property
    def options(self) -> Options:
        headers = dict(self.adapter.request_headers())
        params = self.adapter.request_params()

        options = {
            header_to_opt(k): v for k, v in headers.items() if k.startswith("X-Up-")
        }
        options.update(
            {param_to_opt(k): v for k, v in params.items() if k.startswith("_up_")}
        )
        return Options.parse(options, self.adapter)

    def __bool__(self) -> bool:
        """Returns true if the request is triggered via Unpoly.

        This basically checks if the `X-Up-Version` header is set."""
        return bool(self.options.version)

    def set_title(self, value: str):
        """Sets the title so Unpoly can update the <title> tag (`X-Up-Title`)."""
        self.options.title = value

    def emit(self, type: str, options: Mapping[str, Any]):
        """Emit events to the frontend (`X-Up-Events`)."""
        if not self.options.events:
            self.options.events = []
        self.options.events.append(dict(type=type, **options))

    @cached_property
    def cache(self) -> Cache:
        """Access to the :class:`.Cache` functionality"""
        return Cache(self)

    @property
    def version(self) -> str:
        """Unpoly version (`X-Up-Version`)."""
        return self.options.version

    @property
    def target(self) -> str:
        """Gives access to the request's target (`X-Up-Target`).
        If the server set a new target it will return that instead."""
        return self.options.server_target or self.options.target

    @target.setter
    def target(self, new_target: str):
        self.options.server_target = new_target

    @property
    def mode(self) -> str:
        """Returns the request's mode (`X-Up-Mode`)."""
        return self.options.mode

    @property
    def context(self) -> Mapping[str, Any]:
        """Returns the current context.
        Initially this is `X-Up-Context` but the server can modify the
        returned dictionary to update the values on the client."""
        return self.options.context

    @cached_property
    def layer(self) -> Layer:
        """Access the current :class:`.Layer` configuration."""
        return Layer(self, self.mode, self.context)

    @cached_property
    def reload_from_time(self) -> Optional[datetime]:
        """The time the fragment was last updated (`X-Up-Reload-From-Time`)."""
        return self.options.reload_from_time

    @property
    def validate(self) -> Optional[str]:
        """Returns the field Unpoly is trying to validate (`X-Up-Validate`)."""
        return self.options.validate

    @property
    def fail_target(self) -> str:
        """Gives access to the request's failure target (`X-Up-Fail-Target`).
        If the server set a new target it will return that instead."""
        return self.options.server_target or self.options.fail_target

    @property
    def fail_mode(self) -> str:
        """Returns the request's failure mode (`X-Up-Fail-Mode`)."""
        return self.options.fail_mode

    @property
    def fail_context(self) -> str:
        """Returns the current failure context (`X-Up-Fail-Context`)."""
        return self.options.fail_context

    @cached_property
    def fail_layer(self) -> Layer:
        """Access the current :class:`.Layer` configuration for failures."""
        return Layer(self, self.fail_mode, self.fail_context)

    @property
    def needs_cookie(self):
        """Checks whether the response should set the `_up_method` cookie."""
        return self.adapter.method != "GET" and not bool(self)

    def finalize_response(self, response):
        """Finalize the response by settings required headers & cookies.

        It should be noted that the response is passed as is to current adapter, which
        knows how to set headers on the response etc.
        """
        self.adapter.set_cookie(response, self.needs_cookie)

        if not self:
            return

        redirect_uri = self.adapter.redirect_uri(response)
        serialized_options = self.options.serialize(self.adapter)
        # Handle redirects
        if redirect_uri:
            if "context" in serialized_options:
                serialized_options["context_diff"] = serialized_options.pop("context")
            params = {opt_to_param(k): v for k, v in serialized_options.items()}
            sep = "&" if "?" in redirect_uri else "?"
            if params:
                redirect_uri += sep + urlencode(params)
            self.adapter.set_redirect_uri(response, redirect_uri)
        else:
            loc = self.adapter.location
            if "?" in loc and "_up_" in loc:  # Not 100% exact, but will do
                loc, qs = loc.split("?", 1)
                qs = {k: v for k, v in parse_qs(qs).items() if not k.startswith("_up_")}
                if qs:
                    loc = f"{loc}?{urlencode(qs, doseq=True)}"
            serialized_options["location"] = loc
            serialized_options["method"] = self.adapter.method
            headers = {opt_to_header(k): v for k, v in serialized_options.items()}
            self.adapter.set_headers(response, headers)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['unpoly', 'unpoly.contrib']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.2.0,<22.0.0']

extras_require = \
{':python_version < "3.8"': ['backports.cached_property>=1.0.1,<2.0.0']}

setup_kwargs = {
    'name': 'unpoly',
    'version': '0.2.0',
    'description': 'Framework agnostic implementation of the unpoly server-protocol.',
    'long_description': '[![version](https://img.shields.io/pypi/v/unpoly.svg)](https://pypi.org/project/unpoly)\n[![python versions](https://img.shields.io/pypi/pyversions/unpoly.svg)](https://pypi.org/project/unpoly)\n[![docs](https://img.shields.io/readthedocs/unpoly)](https://unpoly.readthedocs.io)\n[![pipeline status](https://gitlab.com/rocketduck/python-unpoly/badges/main/pipeline.svg)](https://gitlab.com/rocketduck/python-unpoly/-/commits/main)\n[![coverage report](https://gitlab.com/rocketduck/python-unpoly/badges/main/coverage.svg)](https://gitlab.com/rocketduck/python-unpoly/-/commits/main) \n\n# Unpoly\n\nUnpoly is a framework agnostic python library implementing the [Unpoly server protocol](https://unpoly.com/up.protocol).\n\n## Features\n\n* **Full protocol implementation**: The whole Unpoly server protocol is implemented and well tested.\n* **Django support**: Out of the box we currently ship a middleware for Django support.\n* **Easily extendable**: The library abstracts the actual HTTP stuff via adapters and can easily plugged into frameworks like Flask etc.\n\n## Download & Install\n\n```\npip install unpoly\n```\n\n### Usage with Django\n\nAdd `unpoly.contrib.django.UnpolyMiddleware` to your middlewares and then you can access `request.up`. Details can be found in the usage section of the [docs](https://unpoly.readthedocs.io/usage.html).\n\nExample usage:\n\n```py\ndef my_view(request):\n    if request.up: # Unpoly request\n        # Send an event down to unpoly\n        request.up.emit("test:event", {"event": "params"})\n        # ... and also clear the cache for certain paths\n        request.up.clear("/users/*")\n    else:\n        ...\n\ndef form_view(request):\n    form = MyForm(request.GET)\n    # When unpoly wants to validate a form it sends\n    # along X-Up-Validate which contains the field\n    # being validated.\n    if form.is_valid() and not request.up.validate:\n        form.save()\n    return render(request, "template.html", {"form": form})\n```\n\n### Usage with Flask etc\n\nSubclass `unpoly.adapter.BaseAdapter` and initialize `unpoly.Unpoly` with it for every request (see the [docs](https://unpoly.readthedocs.io/adapters.html) for details).',
    'author': 'Florian Apolloner',
    'author_email': 'florian@apolloner.eu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/rocketduck/python-unpoly',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<3.11',
}


setup(**setup_kwargs)



from unpoly import Unpoly
from unpoly.adapter import BaseAdapter

UP_METHOD_COOKIE = "_up_method"


class UnpolyMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        request.up = up = Unpoly(DjangoAdapter(request))
        response = self.get_response(request)
        up.finalize_response(response)
        return response


class DjangoAdapter(BaseAdapter):
    def __init__(self, request):
        self.request = request

    def request_headers(self):
        return self.request.headers

    def request_params(self):
        return self.request.GET

    def redirect_uri(self, response):
        return (
            response.headers.get("Location")
            if 300 <= response.status_code < 400
            else False
        )

    def set_redirect_uri(self, response, uri):
        response.headers["Location"] = uri

    def set_headers(self, response, headers):
        for k, v in headers.items():
            response.headers[k] = v

    def set_cookie(self, response, needs_cookie=False):
        if needs_cookie:
            response.set_cookie(UP_METHOD_COOKIE, self.method)
        elif UP_METHOD_COOKIE in self.request.COOKIES:
            response.delete_cookie(UP_METHOD_COOKIE)

    @property
    def method(self):
        return self.request.method

    @property
    def location(self):
        return self.request.get_full_path_info()
